﻿using SimpleJSON;
using System;
using System.Collections.Generic;

namespace AwasAutoJuice {

    internal class AwasActionConfigurationRepository {

        private Dictionary<string, AwasActionConfiguration> actionsConfig = new Dictionary<string, AwasActionConfiguration>();
        public readonly List<string> actionNames;

        public AwasActionConfigurationRepository(bool isFemale) {
            Logger.Debug("AwasActionConfigurationRepository ctor");
            actionNames = isFemale ? FemaleActionNames.names : MaleActionNames.names;
            foreach (string actionName in actionNames) {
                Add(actionName, new AwasActionConfiguration(actionName));
            }
        }

        public void RestoreFromJSON(JSONArray actionsConfigJson) {
            Logger.Debug("AwasActionConfigurationRepository RestoreFromJSON");
            for (int i = 0; i < actionsConfigJson.Count; ++i) {
                JSONClass actionConfigJson = actionsConfigJson[i].AsObject;
                string actionName = actionConfigJson["name"];
                AwasActionConfiguration actionConfig = new AwasActionConfiguration(actionName);
                actionConfig.RestoreFromJSON(actionConfigJson);
                Add(actionName, actionConfig);
            }
        }

        public JSONArray GetJSON() {
            JSONArray actionsConfigJson = new JSONArray();
            foreach (KeyValuePair<string, AwasActionConfiguration> pair in actionsConfig) {
                actionsConfigJson.Add(pair.Value.GetJSON());
            }
            return actionsConfigJson;
        }

        public AwasActionConfiguration Get(string name) {
            return actionsConfig[name];
        }

        public void RestoreFactoryDefaults() {
            Logger.Debug("AwasActionConfigurationRepository RestoreFactoryDefaults");
            foreach (string actionName in actionNames) {
                Get(actionName).UpdateSettings(AwasActionDefaults.Get(actionName));
            }
        }

        private void Add(string name, AwasActionConfiguration config) {
            if (actionsConfig.ContainsKey(name)) {
                actionsConfig[name] = config;
                return;
            }
            actionsConfig.Add(name, config);
        }
    }

    internal class AwasActionConfiguration {
        private string awasActionName;
        private AwasActionSettings settings;
        public readonly AwasActionConfigStorables ui;

        public AwasActionConfiguration(string actionName) {
            settings = AwasActionDefaults.Get(actionName);
            awasActionName = actionName;
            ui = new AwasActionConfigStorables(this);
        }

        public string displayName { get { return settings.displayName; } }
        public string name { get { return awasActionName; } }
        public bool isDrip { get { return settings.isDrip; } }
        public bool isSim { get { return settings.isSim; } }
        public float maxDuration { get { return settings.maxDuration; } set { settings.maxDuration = value; } }
        public float visibleAlpha { get { return settings.visibleAlpha; } set { settings.visibleAlpha = value; } }
        public float friction { get { return settings.friction; } set { settings.friction = value; } }
        public float dripFriction { get { return settings.dripFriction; } set { settings.dripFriction = value; } }
        public float compression { get { return settings.compression; } set { settings.compression = value; } }
        public float dripCompression { get { return settings.dripCompression; } set { settings.dripCompression = value; } }
        public float size { get { return settings.size; } set { settings.size = value; } }
        public float dripSize { get { return settings.dripSize; } set { settings.dripSize = value; } }
        public float stiffness { get { return settings.stiffness; } set { settings.stiffness = value; } }
        public float dripStiffness { get { return settings.dripStiffness; } set { settings.dripStiffness = value; } }

        // Advanced settings
        public float jointStrength { get { return settings.jointStrength; } set { settings.jointStrength = value; } }
        public float dripJointStrength { get { return settings.dripJointStrength; } set { settings.dripJointStrength = value; } }
        public float dripCollisionRadius { get { return settings.dripCollisionRadius; } set { settings.dripCollisionRadius = value; } }
        public float dripCollisionPower { get { return settings.dripCollisionPower; } set { settings.dripCollisionPower = value; } }
        
        // Don't mess with these
        public float skinScale { get { return settings.skinScale; } set { settings.skinScale = value; } }
        public float dripSkinScale { get { return settings.dripSkinScale; } set { settings.dripSkinScale = value; } }

        public void UpdateSettings(AwasActionSettings newSettings) {
            settings = newSettings;
        }

        public void RestoreFromJSON(JSONClass actionJsonConfig) {
            try {
                settings.maxDuration     = ParseAndValidateSetting(actionJsonConfig[SettingNames.maxDuration], ActionSettingsRange.duration);
                settings.visibleAlpha    = ParseAndValidateSetting(actionJsonConfig[SettingNames.visibleAlpha], ActionSettingsRange.alpha);
                settings.friction        = ParseAndValidateSetting(actionJsonConfig[SettingNames.friction], ActionSettingsRange.zeroToOne);
                settings.dripFriction    = ParseAndValidateSetting(actionJsonConfig[SettingNames.dripFriction], ActionSettingsRange.zeroToOne);
                settings.dripCompression = ParseAndValidateSetting(actionJsonConfig[SettingNames.dripCompression], ActionSettingsRange.nonZero);
                settings.size            = ParseAndValidateSetting(actionJsonConfig[SettingNames.size], ActionSettingsRange.size);
                settings.dripSize        = ParseAndValidateSetting(actionJsonConfig[SettingNames.dripSize], ActionSettingsRange.size);
                settings.stiffness       = ParseAndValidateSetting(actionJsonConfig[SettingNames.stiffness], ActionSettingsRange.zeroToOne);
                settings.dripStiffness   = ParseAndValidateSetting(actionJsonConfig[SettingNames.dripStiffness], ActionSettingsRange.zeroToOne);
                settings.compression     = ParseAndValidateSetting(actionJsonConfig[SettingNames.compression], ActionSettingsRange.nonZero);
                settings.jointStrength   = ParseAndValidateSetting(actionJsonConfig[SettingNames.jointStrength], ActionSettingsRange.zeroToOne);
                settings.dripJointStrength = ParseAndValidateSetting(actionJsonConfig[SettingNames.dripJointStrength], ActionSettingsRange.zeroToOne);
                settings.dripCollisionRadius = ParseAndValidateSetting(actionJsonConfig[SettingNames.dripCollisionRadius], ActionSettingsRange.nonZero);
                settings.dripCollisionPower = ParseAndValidateSetting(actionJsonConfig[SettingNames.dripCollisionPower], ActionSettingsRange.zeroToOne);
                settings.skinScale       = ParseAndValidateSetting(actionJsonConfig[SettingNames.skinScale], ActionSettingsRange.size);
                settings.dripSkinScale   = ParseAndValidateSetting(actionJsonConfig[SettingNames.dripSkinScale], ActionSettingsRange.size);
            } catch (Exception e) {
                SuperController.LogError(e.Message);
            }
        }

        public JSONClass GetJSON() {
            return new JSONClass {
                ["name"] = awasActionName,
                [SettingNames.maxDuration]     = settings.maxDuration.ToString(),
                [SettingNames.visibleAlpha]    = settings.visibleAlpha.ToString(),
                [SettingNames.friction]        = settings.friction.ToString(),
                [SettingNames.dripFriction]    = settings.dripFriction.ToString(),
                [SettingNames.dripCompression] = settings.dripCompression.ToString(),
                [SettingNames.size]            = settings.size.ToString(),
                [SettingNames.dripSize]        = settings.dripSize.ToString(),
                [SettingNames.stiffness]       = settings.stiffness.ToString(),
                [SettingNames.dripStiffness]   = settings.dripStiffness.ToString(),
                [SettingNames.compression]     = settings.compression.ToString(),
                [SettingNames.jointStrength]   = settings.jointStrength.ToString(),
                [SettingNames.dripJointStrength] = settings.dripJointStrength.ToString(),
                [SettingNames.dripCollisionRadius] = settings.dripCollisionRadius.ToString(),
                [SettingNames.dripCollisionPower] = settings.dripCollisionPower.ToString(),
                [SettingNames.skinScale]       = settings.skinScale.ToString(),
                [SettingNames.dripSkinScale]   = settings.dripSkinScale.ToString(),

            };
        }

        private float ParseAndValidateSetting(string value, ActionSettingsRange.Range range) {
            float parsedValue = float.Parse(value);
            if (parsedValue < range.min) parsedValue = range.min;
            if (parsedValue > range.max) parsedValue = range.max;
            return parsedValue;
        }
    }

    internal static class FemaleActionNames {
        public static readonly string PussyWet = "PussyWet";
        public static readonly string PussyWetJuicy = "PussyWetJuicy";
        public static readonly string PussyFullOfCum = "PussyFullOfCum";
        public static readonly string PussyWetDrip = "PussyWetDrip";
        public static readonly string PussyCumDrip = "PussyCumDrip";
        public static readonly string MouthFullOfCum = "MouthFullOfCum";
        public static readonly string MouthCumDrip = "MouthCumDrip";
        public static readonly string MouthCumSticky = "MouthCumSticky";
        public static readonly string NeckCumDrip = "NeckCumDrip";
        public static readonly List<string> names = new List<string>() { PussyWet, PussyWetJuicy, PussyFullOfCum, PussyWetDrip, PussyCumDrip, MouthFullOfCum, MouthCumDrip, MouthCumSticky, NeckCumDrip };
    }

    internal static class MaleActionNames {
        public static readonly string PenisWet = "PenisWet";
        public static readonly string PenisWetTip = "PenisWetTip";
        public static readonly string PenisCum = "PenisCum";
        public static readonly string Jizz = "Jizz";
        public static readonly string JizzSticky = "JizzSticky";
        public static readonly string JizzOnMouth = "JizzOnMouth";
        public static readonly string JizzInMouth = "JizzInMouth";
        public static readonly List<string> names = new List<string>() { PenisWet, PenisWetTip, PenisCum, Jizz, JizzSticky, JizzOnMouth, JizzInMouth };
    }

    internal static class SettingNames {
        public static readonly string maxDuration = "maxDuration";
        public static readonly string visibleAlpha = "visibleAlpha";
        public static readonly string friction = "friction";
        public static readonly string dripFriction = "dripFriction";
        public static readonly string dripCompression = "dripCompression";
        public static readonly string size = "size";
        public static readonly string dripSize= "dripSize";
        public static readonly string stiffness = "stiffness";
        public static readonly string dripStiffness = "dripStiffness";
        public static readonly string compression = "compression";
        public static readonly string jointStrength = "jointStrength";
        public static readonly string dripJointStrength = "dripJointStrength";
        public static readonly string dripCollisionRadius = "dripCollisionRadius";
        public static readonly string dripCollisionPower = "dripCollisionPower";
        public static readonly string skinScale = "skinScale";
        public static readonly string dripSkinScale = "dripSkinScale";
    }

    internal struct AwasActionSettings {
        public bool isDrip;
        public bool isSim;
        public string displayName;
        public float maxDuration;
        public float visibleAlpha;
        public float friction;
        public float dripFriction;
        public float compression;
        public float dripCompression;
        public float size;
        public float dripSize;
        public float stiffness;
        public float dripStiffness;
        public float jointStrength;
        public float dripJointStrength;
        public float dripCollisionRadius;
        public float dripCollisionPower;
        public float skinScale;
        public float dripSkinScale;
    }

    internal static class ActionSettingsRange {
        public class Range { public float min; public float max; }
        public static readonly Range duration = new Range() { min = 3f, max = 10f };
        public static readonly Range alpha = new Range() { min = -1f, max = 1f };
        public static readonly Range size = new Range() { min = 0.1f, max = 2f };
        public static readonly Range zeroToOne = new Range() { min = 0f, max = 1f };
        public static readonly Range nonZero = new Range() { min = 0.001f, max = 1f };
    }

    internal static class AwasActionDefaults {
        public static AwasActionSettings Get(string name) {
            return settings[name];
        }
        private static readonly Dictionary<string, AwasActionSettings> settings = new Dictionary<string, AwasActionSettings>() {
            { FemaleActionNames.PussyWet, new AwasActionSettings() {
                isDrip = false,
                isSim = false,
                displayName = "Wet Pussy",
                visibleAlpha = -0.1f
            }},
            { FemaleActionNames.PussyWetJuicy, new AwasActionSettings() {
                isDrip = false,
                isSim = true,
                displayName = "Wet Juicy Pussy",
                maxDuration = 0f,
                visibleAlpha = -0.1f,
                friction = 1f,
                dripFriction = 0f,
                size = 2f,
                dripSize = 0f,
                stiffness = 0.08f,
                dripStiffness = 0f,
                compression = 0f,
                dripCompression = 0f,
                jointStrength = 0.2f,
                dripJointStrength = 0f,
                dripCollisionRadius = 0.001f,
                dripCollisionPower = 0.1f,
                skinScale = 1f,
                dripSkinScale = 1f
            }},
            { FemaleActionNames.PussyFullOfCum, new AwasActionSettings() {
                isDrip = false,
                isSim = true,
                displayName = "Pussy Full of Cum",
                maxDuration = 0f,
                visibleAlpha = -0.01f,
                friction = 1f,
                dripFriction = 0f,
                dripCompression = 0f,
                size = 2f,
                dripSize = 0f,
                stiffness = 0.03f,
                dripStiffness = 0f,
                compression = 0.001f,
                jointStrength = 0.15f,
                dripJointStrength = 0f,
                dripCollisionRadius = 0.001f,
                dripCollisionPower = 0.2f,
                skinScale = 1f,
                dripSkinScale = 1f
            }},
            { FemaleActionNames.PussyWetDrip, new AwasActionSettings() {    
                isDrip = true,
                isSim = true,
                displayName = "Wet Pussy Drip",
                maxDuration = 8f,
                visibleAlpha = -0.2f,
                friction = 1f,
                dripFriction = 0.3f,
                size = 0.8f,
                dripSize = 0.5f,
                stiffness = 0.4f,
                dripStiffness = 0.040f,
                compression = 0.3f,
                dripCompression = 0.2f,
                jointStrength = 0.3f,
                dripJointStrength = 0.01f,
                dripCollisionRadius = 0.005f,
                dripCollisionPower = 0.2f,
                skinScale = 1f,
                dripSkinScale = 1f,
            }},
            { FemaleActionNames.PussyCumDrip, new AwasActionSettings() {
                isDrip = true,
                isSim = true,
                displayName = "Pussy Cum Drip",
                maxDuration = 8f,
                visibleAlpha = -0.2f,
                friction = 0.8f,
                dripFriction = 0.4f,
                size = 1f,
                dripSize = 0.7f,
                stiffness = 0.001f,
                dripStiffness = 0.065f,
                compression = 0.5f,
                dripCompression = 0.5f,
                jointStrength = 0.5f,
                dripJointStrength = 0.01f,
                dripCollisionRadius = 0.005f,
                dripCollisionPower = 0.2f,
                skinScale = 1f,
                dripSkinScale = 1f,
            }},
            { FemaleActionNames.MouthFullOfCum, new AwasActionSettings() {
                isDrip = false,
                isSim = true,
                displayName = "Mouth Full of Cum",
                maxDuration = 0f,
                visibleAlpha = 0.0f,
                friction = 1f,
                dripFriction = 0f,
                size = 0.8f,
                dripSize = 0f,
                stiffness = 0.005f,
                dripStiffness = 0f,
                compression = 0f,
                jointStrength = 0.3f,
                dripCollisionPower = 0.2f,
                dripCollisionRadius = 0.001f
            }},
            { FemaleActionNames.MouthCumDrip, new AwasActionSettings() {
                isDrip = true,
                isSim = true,
                displayName = "Mouth Cum Drip",
                maxDuration = 8f,
                visibleAlpha = -0.3f,
                friction = 0.8f,
                dripFriction = 0.4f,
                compression = 0.5f,
                dripCompression = 0.5f,
                size = 0.5f,
                dripSize = 1.4f,
                stiffness = 0.3f,
                dripStiffness = 0.015f,
                jointStrength = 0.5f,
                dripJointStrength = 0.1f,
                dripCollisionRadius = 0.002f,
                dripCollisionPower = 1f,
            }},
            { FemaleActionNames.MouthCumSticky, new AwasActionSettings() {
                isDrip = false,
                isSim = true,
                displayName = "Sticky Mouth Cum",
                maxDuration = 0f,
                visibleAlpha = -0.3f,
                friction = 0.2f,
                dripFriction = 0f,
                compression = 0f,
                dripCompression = 0f,
                size = 0.5f,
                dripSize = 0f,
                stiffness = 0.05f,
                dripStiffness = 0f,
                jointStrength = 0.4f,
                dripCollisionRadius = 0.001f,
                dripCollisionPower = 0.8f,
            }},
            { FemaleActionNames.NeckCumDrip, new AwasActionSettings() {
                isDrip = true,
                isSim = true,
                displayName = "Cum Neck Drip",
                maxDuration = 8f,
                visibleAlpha = -0.4f,
                friction = 0.7f,
                dripFriction = 0.3f,
                dripCompression = 1f,
                size = 1f,
                dripSize = 0.7f,
                stiffness = 0.2f,
                dripStiffness = 0.07f,
                compression = 1f,
                jointStrength = 0.1f,
                dripJointStrength = 0.06f,
                dripCollisionRadius = 0.002f,
                dripCollisionPower = 0.6f
            }},
            { MaleActionNames.PenisWet, new AwasActionSettings() {
                isDrip = false,
                isSim = false,
                displayName = "Wet Penis",
                visibleAlpha = -0.8f

            }},
            { MaleActionNames.PenisWetTip, new AwasActionSettings() {
                isDrip = false,
                isSim = false,
                displayName = "Wet Penis Tip",
                visibleAlpha = -0.6f

            }},
            { MaleActionNames.PenisCum, new AwasActionSettings() {
                isDrip = false,
                isSim = false,
                displayName = "Penis Covered in Cum",
                visibleAlpha = -0.75f
            }},
            { MaleActionNames.Jizz, new AwasActionSettings() {
                isDrip = true,
                isSim = true,
                displayName = "Jizz",
                maxDuration = 8f,
                visibleAlpha = -0.2f,
                friction = 0.8f,
                dripFriction = 0.5f,
                size = 1f,
                dripSize = 1f,
                stiffness = 0.5f,
                dripStiffness = 0.17f,
                compression = 0.8f,
                dripCompression = 0.5f,
                jointStrength = 1f,
                dripCollisionRadius = 0.001f,
                dripCollisionPower = 0.5f,
                skinScale = 2f,
                dripSkinScale = 1f
            }},
            { MaleActionNames.JizzSticky, new AwasActionSettings() {
                isDrip = true,
                isSim = true,
                displayName = "Sticky Jizz",
                maxDuration = 8f,
                visibleAlpha = -0.2f,
                friction = 0.2f,
                dripFriction = 0.3f,
                compression = 0.8f,
                dripCompression = 0.8f,
                size = 1f,
                dripSize = 1f,
                stiffness = 0.5f,
                dripStiffness = 0.07f,
                jointStrength = 1f,
                dripJointStrength = 0f,
                dripCollisionRadius = 0.001f,
                dripCollisionPower = 0.5f,
                skinScale = 2f,
                dripSkinScale = 2f
            }},
            { MaleActionNames.JizzOnMouth, new AwasActionSettings() {
                isDrip = true,
                isSim = true,
                displayName = "Jizz on Mouth",
                maxDuration = 8f,
                visibleAlpha = -0.2f,
                friction = 0.6f,
                dripFriction = 0.8f,
                size = 1f,
                dripSize = 1f,
                stiffness = 0.5f,
                dripStiffness = 0.2f,
                jointStrength = 1f,
                dripJointStrength = 0.07f,
                compression = 0.5f,
                dripCompression = 0.5f,
                dripCollisionRadius = 0.001f,
                dripCollisionPower = 0.5f,
                skinScale = 0.7f,
                dripSkinScale = 1f
            }},
            { MaleActionNames.JizzInMouth, new AwasActionSettings() {
                isDrip = true,
                isSim = true,
                displayName = "Jizz in Mouth",
                maxDuration = 8f,
                visibleAlpha = -0.2f,
                friction = 0.7f,
                dripFriction = 0.7f,
                compression = 0.7f,
                dripCompression = 0.5f,
                size = 1f,
                dripSize = 1f,
                stiffness = 0.8f,
                dripStiffness = 0.6f,
                skinScale = 0.1f,
                jointStrength = 1f,
                dripJointStrength = 0.07f,
                dripCollisionRadius = 0.005f,
                dripCollisionPower = 0.8f
            }},
        };
    }
}
