﻿using System;
using System.Collections;
using UnityEngine;

namespace AwasAutoJuice {
    
    internal class AwasFluidAction {
        
        private AwasFluidAtomFactory atomFactory;
        protected readonly AwasFluidCloth afcloth;
        protected readonly float notVisibleAlphaTarget = -1f;
        protected readonly AwasActionConfiguration actionConfig;
        protected readonly float visibleAlphaTarget;
        protected float fadeInSeconds = 1.5f;
        protected float fadeOutSeconds = 1.5f;
        protected float fluidPrepDuration = 1.5f;
        protected float maxDistanceScaleOverride = 1f;
        protected bool actionStarted = false;
        protected Logger logger;

        public AwasFluidAction(AwasFluidCloth awasFluidCloth, AwasActionConfiguration actionConfig) {
            if (awasFluidCloth == null) {
                throw new ArgumentNullException("awasFluidCloth");
            }
            afcloth = awasFluidCloth;
            atomFactory = new AwasFluidAtomFactory(afcloth.GetContainingAtom().name, afcloth.name);
            this.actionConfig = actionConfig;
            visibleAlphaTarget = actionConfig.visibleAlpha;
            maxDistanceScaleOverride = actionConfig.dripSize;
            logger = new Logger($"{actionConfig.name} {afcloth.name}");
            try {
                logger.ContextDebug($"instance created with configured collision radius {actionConfig.dripCollisionRadius}");
            } catch (Exception e) {
                logger.ContextDebug(e.Message);
            }
        }

        public virtual void DoAction() {
            DoActionIfClothReady(Start());
        }

        public virtual void DoAction(FreeControllerV3 controller) {
            DoActionIfClothReady(Start(controller));
        }

        public virtual void StopAction() {
            actionStarted = false;
            StartCoroutine(Stop());
        }

        private bool ClothIsReady() {
            return !afcloth.IsActive();
        }

        private void DoActionIfClothReady(IEnumerator startAction) {
            if (ClothIsReady()) {
                actionStarted = true;
                //logger.ContextDebug("DoAction base; activate, set alpha -1");
                afcloth.ResetPhysics(); 
                afcloth.Reset();
                afcloth.SetActive(true);
                afcloth.SetClothActive(true);
                StartCoroutine(startAction);
                StartCoroutine(StopActionIfDrip());
            }
        }

        private IEnumerator StopActionIfDrip() {
            if (actionConfig.isDrip) {
                yield return new WaitForSecondsRealtime(fluidPrepDuration + actionConfig.maxDuration);
                //logger.ContextDebug($"Duration of {actionConfig.maxDuration} reached");
                yield return Stop();
                logger.ContextDebug("Stopped and finished successfully");
            } else {
                yield break;
            }
        }

        protected virtual IEnumerator Start() {
            throw new NotImplementedException();
        }

        protected virtual IEnumerator Start(FreeControllerV3 controller) {
            throw new NotImplementedException();
        }

        protected virtual IEnumerator Stop() {
            //logger.ContextDebug("Base action stop");
            // TODO nasty hack - either remove this bool or do something different around actions running over the top of each other with coroutines
            actionStarted = true;
            yield return LerpAlphaToNotVisible();
            FinishAction();
            yield return null;
            //logger.ContextDebug("Base Stop done");
        }

        protected virtual void FinishAction() {
            //logger.ContextDebug("FinishAction base; reattach, reset and deactivate");
            actionStarted = false;
            afcloth.Reset();
            afcloth.SetClothActive(false);
            afcloth.SetDetached(false);
            afcloth.SetSimEnabled(false);
            afcloth.SetActive(false);
        }

        protected IEnumerator LerpAlphaToVisible(float duration = -1) {
            duration = duration < 0 ? fadeInSeconds : duration;
            return LerpClothSetting(notVisibleAlphaTarget, visibleAlphaTarget, duration, (newValue) => afcloth.SetAlpha(newValue));
        }

        protected IEnumerator LerpAlphaToNotVisible() {
            return LerpClothSetting(visibleAlphaTarget, notVisibleAlphaTarget, fadeOutSeconds, (newValue) => afcloth.SetAlpha(newValue));
        }

        protected IEnumerator LerpFriction(float duration) {
            return LerpClothSetting(actionConfig.friction, actionConfig.dripFriction, duration, (newValue) => afcloth.SetFriction(newValue));
        }

        protected IEnumerator LerpStiffness(float duration) {
            return LerpClothSetting(actionConfig.stiffness, actionConfig.dripStiffness, duration, (newValue) => afcloth.SetStiffness(newValue));
        }

        protected IEnumerator LerpDistanceScale(float duration) {
            return LerpClothSetting(actionConfig.size, maxDistanceScaleOverride, duration, (newValue) => afcloth.SetDistanceScale(newValue));
        }

        protected IEnumerator LerpCompression(float duration) {
            return LerpClothSetting(actionConfig.compression, actionConfig.dripCompression, duration, (newValue) => afcloth.SetCompressionResistance(newValue));
        }

        protected IEnumerator LerpJointStrength(float duration) {
            return LerpClothSetting(actionConfig.jointStrength, actionConfig.dripJointStrength, duration, (newValue) => afcloth.SetJointStrength(newValue));
        }

        protected IEnumerator LerpCollisionPower(float start, float end, float duration) {
            return LerpClothSetting(start, end, duration, (newValue) => afcloth.SetCollisionPower(newValue));
        }

        protected IEnumerator LerpForce(Vector3 start, Vector3 end, float duration) {
            float step = 0.05f;
            float iterationWait = duration * step;
            for (float t = 0; t <= 1; t += step) {
                afcloth.SetForce(Vector3.Lerp(start, end, t));
                if (!actionStarted) yield break;
                yield return new WaitForSecondsRealtime(iterationWait);
            }
            afcloth.SetForce(end);
            yield return null;
        }

        private IEnumerator LerpClothSetting(float startValue, float endValue, float duration, Action<float> clothSetting)  {
            float step = 0.05f;
            float iterationWait = duration * step;
            for (float t = 0; t <= 1; t += step) {
                clothSetting(Mathf.Lerp(startValue, endValue, t));
                if (!actionStarted) yield break;
                yield return new WaitForSecondsRealtime(iterationWait);
            }
            clothSetting(endValue);
            yield return null;
        }

        protected IEnumerator WaitForEndOfFrame() {
            yield return new WaitForEndOfFrame();
        }

        protected IEnumerator AddNewClothGrabSphere(Action<AwasFluidGrabSphere> assignAtom) {
            yield return atomFactory.CreateNewGrabSphere(assignAtom);
        }

        protected IEnumerator WaitForFixedUpdateFrames(int frames = 1) {
            for (int i = 0; i <= frames; ++i) {
                yield return new WaitForFixedUpdate();
            }
        }

        protected Coroutine StartCoroutine(IEnumerator enumerator) {
            return SuperController.singleton.StartCoroutine(enumerator);
        }

        protected IEnumerator DoAfterSeconds(float seconds, Action doThis) {
            yield return new WaitForSecondsRealtime(seconds);
            doThis();
        }
    }
    
}
