﻿using System.Collections.Generic;
using System.Collections;
using System.Linq;
using UnityEngine;

namespace AwasAutoJuice {

    internal class AwasFluidBodyTriggerService {

        private readonly Atom containingAtom;
        private readonly AwasFluidActionService actionService;
        private readonly Logger logger;
        private AwasFluidAtomBodyState bodyState;
        private List<AtomRigidBodyCollisionTrigger> triggers = new List<AtomRigidBodyCollisionTrigger>();
        private bool isThroatCumSequenceStarted = false;

        public AwasFluidBodyTriggerService(Atom containingAtom, bool isFemale, AwasFluidActionService actionService) {
            this.containingAtom = containingAtom;
            this.actionService = actionService;
            bodyState = new AwasFluidAtomBodyState(containingAtom.name);

            logger = new Logger($"{containingAtom.name} TriggerService");
            
            if (isFemale) {
                SetFemaleTriggers();
            } else {
                SetMaleTriggers(); 
            }
        }

        public void SetActive(bool active) {
            triggers.ForEach(trigger => trigger.SetEnabled(active));
            if (active) {
                bodyState.ResetBodyState();
                bodyState.shouldUpdate = true;
                bodyState.Monitor();
                triggers.ForEach(trigger => trigger.Debug(RunMode.isDev));
            } else {
                bodyState.shouldUpdate = false;
                triggers.ForEach(trigger => trigger.Debug(false));
            }
        }

        public void RemoveAllTriggersAndHandlers() {
            logger.ContextDebug("RemoveAllTriggersAndHandlers");
            triggers.ForEach(trigger => {
                trigger.Remove();
            });
        }

        private void SetFemaleTriggers() {
            List<string> collisionWhitelist = new List<string>() { "Gen1", "Gen2", "Gen3", "_Collider" };
            List<string> mouthCollisionWhitelist = new List<string>() { "Gen1", "Gen2", "Gen3" };

            FreeControllerV3 pelvisController = containingAtom.freeControllers.First(rb => rb.name == "pelvisControl");
            Rigidbody pelvisRb = containingAtom.rigidbodies.First(rb => rb.name == "pelvis");
            Rigidbody throatTriggerRb = containingAtom.rigidbodies.First(rb => rb.name == "ThroatTrigger");
            Rigidbody lipTriggerRb = containingAtom.rigidbodies.First(rb => rb.name == "LipTrigger");
            Rigidbody mouthTriggerRb = containingAtom.rigidbodies.First(rb => rb.name == "MouthTrigger");
            Rigidbody lNippleTriggerRb = containingAtom.rigidbodies.First(rb => rb.name == "lNippleTrigger");
            Rigidbody rNippleTriggerRb = containingAtom.rigidbodies.First(rb => rb.name == "rNippleTrigger");
            Rigidbody vaginaTriggerRb = containingAtom.rigidbodies.First(rb => rb.name == "VaginaTrigger");
            Rigidbody labiaTriggerRb = containingAtom.rigidbodies.First(rb => rb.name == "LabiaTrigger");

            bodyState.OnAroused(() => actionService.WetPussyActionGetter().DoAction());
            bodyState.OnRandomStimulated(() => actionService.PussyDripActionGetter().DoAction());
            bodyState.OnVeryWetPussy(() => actionService.WetPussyJuicyActionGetter().DoAction());
            bodyState.OnPussyNotQuiteSoWetAnymore(() => actionService.WetPussyJuicyActionGetter().StopAction());
            bodyState.OnThroatMouthCumSequence(() => StartCoroutine(ThroatMouthCumSequence()));
            bodyState.OnLipsCumSequence(() => StartCoroutine(LipsMouthCumSequence()));
            bodyState.OnPussyCumSequence(() => StartCoroutine(PussyCumSequence()));
            bodyState.OnPussyCumDrip(() => actionService.PussyCumDripActionGetter().DoAction());
            bodyState.OnPussyDrip(() => actionService.PussyDripActionGetter().DoAction());

            //LabiaTrigger
            //VaginaTrigger
            //DeepVaginaTrigger
            //DeeperVaginaTrigger
            //lNippleTrigger
            //rNippleTrigger
            //LipTrigger
            //MouthTrigger
            //ThroatTrigger

            AtomRigidBodyCollisionTrigger clitorisTrigger = CreateNewTrigger(0.03f, 1f, pelvisRb, (pelvisRb.transform.forward * 0.04f) + (pelvisRb.transform.up * -0.18f));
            clitorisTrigger.OnTrigger(true, collisionWhitelist, (Collider collider) => {
                //logger.ContextDebug($"Clitoris collision {collider.name}");
                bodyState.StimulusInput(4);
            });

            AtomRigidBodyCollisionTrigger lipsTrigger = CreateNewTrigger(0.05f, 2f, lipTriggerRb, lipTriggerRb.transform.forward * 0.02f);
            lipsTrigger.OnTrigger(false, mouthCollisionWhitelist, (Collider collider) => {
                bodyState.StimulusInput(2);
                if (collider.name == "JizzBlip") {
                    Logger.Debug($"lips jizz collision {collider.name}");
                    bodyState.LipsCumSequence();
                }
            });

            AtomRigidBodyCollisionTrigger mouthTrigger = CreateNewTrigger(0.05f, 2f, mouthTriggerRb, Vector3.zero);
            mouthTrigger.OnTrigger(false, mouthCollisionWhitelist, (Collider collider) => {
                bodyState.StimulusInput(2);
                if (collider.name == "JizzBlip") {
                    Logger.Debug($"mouth jizz collision {collider.name}");
                    bodyState.ThroatMouthCumSequence();
                }
            });

            AtomRigidBodyCollisionTrigger throatJizzTrigger = CreateNewTrigger(0.05f, 3f, throatTriggerRb, Vector3.zero);
            throatJizzTrigger.OnTrigger(false, mouthCollisionWhitelist, (Collider collider) => {
                bodyState.StimulusInput(3);
                if (collider.name == "JizzBlip") {
                    Logger.Debug($"throat jizz collision {collider.name}");
                    bodyState.ThroatMouthCumSequence();
                }
            });

            AtomRigidBodyCollisionTrigger lNippleTrigger = CreateNewTrigger(0.02f, 2f, lNippleTriggerRb, Vector3.zero);
            lNippleTrigger.OnTrigger(true, collisionWhitelist, (Collider collider) => {
                //logger.ContextDebug($"lnipple collision {collider.name}");
                bodyState.StimulusInput(2);
            });

            AtomRigidBodyCollisionTrigger lBreastTrigger = CreateNewTrigger(0.11f, 2f, lNippleTriggerRb, lNippleTriggerRb.transform.forward * 0.05f);
            lBreastTrigger.OnTrigger(true, (Collider collider) => {
                //logger.ContextDebug($"lBreast {collider.name}");
            });

            AtomRigidBodyCollisionTrigger rBreastTrigger = CreateNewTrigger(0.11f, 2f, rNippleTriggerRb, rNippleTriggerRb.transform.forward * 0.05f);
            rBreastTrigger.OnTrigger(true, (Collider collider) => {
                logger.ContextDebug($"rBreast {collider.name}");
            });

            AtomRigidBodyCollisionTrigger rNippleTrigger = CreateNewTrigger(0.02f, 2f, rNippleTriggerRb, Vector3.zero);
            rNippleTrigger.OnTrigger(true, collisionWhitelist, (Collider collider) => {
                //logger.ContextDebug($"rnipple collision {collider.name}");
                bodyState.StimulusInput(2);
            });

            AtomRigidBodyCollisionTrigger vaginaTrigger = CreateNewTrigger(0.05f, 0.5f, vaginaTriggerRb, Vector3.zero);
            vaginaTrigger.OnTrigger(false, collisionWhitelist, (Collider collider) => {
                bodyState.SetPenetration();
                bodyState.StimulusInput(2);
                //Logger.Debug($"Vagina collision {collider.name}");
                if (collider.name == "JizzBlip") {
                    Logger.Debug($"Vagina jizz collision {collider.name}");
                    bodyState.PussyCumSequence();
                }
            });

            AtomRigidBodyCollisionTrigger labiaTrigger = CreateNewTrigger(0.05f, 0.5f, labiaTriggerRb, Vector3.zero);
            labiaTrigger.OnTrigger(true, collisionWhitelist, (Collider collider) => {
                bodyState.SetPenetration();
                bodyState.StimulusInput(2);
                //Logger.Debug($"Labia collision {collider.name}");
                if (collider.name == "JizzBlip") {
                    Logger.Debug($"Labia jizz collision {collider.name}");
                    bodyState.PussyCumSequence();
                }
            });

            //mouth area Trigger.SetTransformWithOffset(throatTriggerRb.transform, (throatTriggerRb.transform.forward * 0.03f) + (throatTriggerRb.transform.up * -0.01f));

            // Detect jizz / neck collision and trigger neck cum drip.

            // Detect jizz / mouth exterior collision and trigger facial / mouth drips (need to create a cloth item and action for this)

            // Detect jizz / breast collision and trigger left / right breast drips (need to create cloth items and actions for these)
        }

        private IEnumerator ThroatMouthCumSequence() {
            isThroatCumSequenceStarted = true;
            yield return new WaitForSecondsRealtime(0.1f);
            actionService.MouthFullOfCumActionGetter().DoAction();
            yield return new WaitForSecondsRealtime(0.5f);
            isThroatCumSequenceStarted = false;
            actionService.MouthCumDripActionGetter().DoAction();
            yield return new WaitForSecondsRealtime(8f);
            actionService.MouthFullOfCumActionGetter().StopAction();
            actionService.MouthCumDripActionGetter().StopAction();
            actionService.MouthCumDripActionGetter().DoAction();
        }

        private IEnumerator LipsMouthCumSequence() {
            yield return new WaitForSecondsRealtime(0.1f);
            if (isThroatCumSequenceStarted) {
                logger.ContextDebug("LipsMouthCumSequence breakout (throat sequence active?)");
                yield break;
            }
            actionService.MouthCumDripActionGetter().DoAction();
            yield return new WaitForSecondsRealtime(1f);
            actionService.MouthCumStickyActionGetter().DoAction();
            actionService.NeckCumDripActionGetter().DoAction();
            yield return new WaitForSecondsRealtime(8f);
            actionService.MouthCumStickyActionGetter().StopAction();
        }

        private IEnumerator PussyCumSequence() {
            actionService.WetPussyJuicyActionGetter().StopAction();
            yield return new WaitForSecondsRealtime(0.1f);
            actionService.PussyFullOfCumActionGetter().DoAction();
            yield return new WaitForSecondsRealtime(1f);
            actionService.PussyCumDripActionGetter().DoAction();
            yield return new WaitForSecondsRealtime(8f);
            actionService.PussyFullOfCumActionGetter().StopAction();
        }

        private AtomRigidBodyCollisionTrigger CreateNewTrigger(float localScale, float debounceDuration, Rigidbody attachTo, Vector3 offset) {
            AtomRigidBodyCollisionTrigger trigger = new AtomRigidBodyCollisionTrigger(localScale, attachTo, attachTo.name, debounceDuration);
            trigger.SetTransformWithOffset(attachTo.transform, offset);
            triggers.Add(trigger);
            return trigger;
        }

        private Coroutine StartCoroutine(IEnumerator coroutine) {
            return SuperController.singleton.StartCoroutine(coroutine);
        }

        private void SetMaleTriggers() {
            //foreach (var rb in containingAtom.rigidbodies) {
            //    if (rb.name.Contains("Trigger")) logger.ContextDebug($"{rb.name}");
            //    if (rb.name.Contains("Gen")) logger.ContextDebug($"{rb.name}");
            //}
            List<string> penetrationWhitelist = new List<string>() { "LipTrigger", "MouthTrigger", "ThroatTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger" };

            Rigidbody penis1Rb = containingAtom.rigidbodies.First(rb => rb.name == "Gen1");
            Rigidbody penis2Rb = containingAtom.rigidbodies.First(rb => rb.name == "Gen2");
            Rigidbody penis3Rb = containingAtom.rigidbodies.First(rb => rb.name == "Gen3");

            // TODO need a "partially wet" or "wet tip" variation
            bodyState.OnWetPenisTip(() => {
                actionService.PenisWetActionGetter().StopAction();
                actionService.PenisWetTipActionGetter().DoAction();
            });
            bodyState.OnWetPenis(() => {
                actionService.PenisWetTipActionGetter().StopAction();
                actionService.PenisWetActionGetter().DoAction();
            });
            bodyState.OnDryPenis(() => {
                actionService.PenisWetTipActionGetter().StopAction();
                actionService.PenisWetActionGetter().StopAction();
            });
            bodyState.OnCumInMouthOrVagina(() => StartCoroutine(CumInMouthOrVagina()));

            AtomRigidBodyCollisionTrigger penis1Trigger = CreateNewTrigger(0.07f, 0.8f, penis1Rb, Vector3.zero);
            penis1Trigger.OnTrigger(false, penetrationWhitelist, (Collider collider) => {
                logger.ContextDebug($"penis1Trigger collision {collider.name}");
                if (collider.name == "JizzBlip") {
                    bodyState.JizzInput();
                    return;
                }
                bodyState.WetnessInput();
                bodyState.SetPenetration();
            });

            AtomRigidBodyCollisionTrigger penis2Trigger = CreateNewTrigger(0.06f, 0.8f, penis2Rb, Vector3.zero);
            penis2Trigger.OnTrigger(false, penetrationWhitelist, (Collider collider) => {
                logger.ContextDebug($"penis2Trigger collision {collider.name}");
                if (collider.name == "JizzBlip") {
                    bodyState.JizzInput();
                    return;
                }
                bodyState.WetnessInput();
                bodyState.SetPenetration();
            });

            AtomRigidBodyCollisionTrigger penis3Trigger = CreateNewTrigger(0.06f, 0.8f, penis3Rb, Vector3.zero);
            penis3Trigger.OnTrigger(false, penetrationWhitelist, (Collider collider) => {
                logger.ContextDebug($"penis3Trigger collision {collider.name}");
                if (collider.name == "JizzBlip") {
                    return;
                }
                bodyState.TipWetnessInput();
                bodyState.SetPenetration();
            });
        }

        private IEnumerator CumInMouthOrVagina() {
            actionService.PenisWetTipActionGetter().StopAction();
            actionService.PenisWetActionGetter().StopAction();
            actionService.PenisWetCumActionGetter().DoAction();
            yield return new WaitForSecondsRealtime(8f);
            actionService.PenisWetCumActionGetter().StopAction();
        }

        private bool IsValidPenetrationTrigger(string name) {
            return name.Contains("Face") || name.Contains("Lip") || name == "LipTrigger" || name == "MouthTrigger" || name == "ThroatTrigger" || name == "VaginaTrigger" || name == "DeepVaginaTrigger";
        }
    }
}
