﻿using System;
using System.Collections.Generic;
using UnityEngine;

namespace AwasAutoJuice {
    internal class AwasFluidCollisionTrigger {
        private GameObject colliderSphere;
        private Rigidbody colliderRigidBody;
        public STriggerCollide colliderTrigger;

        public AwasFluidCollisionTrigger(float localScale) {
            colliderSphere = GameObject.CreatePrimitive(PrimitiveType.Sphere);
            colliderSphere.name = "JizzCollider";
            colliderSphere.transform.localScale = new Vector3(localScale, localScale, localScale);
            colliderSphere.GetComponent<SphereCollider>().isTrigger = true;
            colliderSphere.GetComponent<SphereCollider>().radius = 1f;

            colliderSphere.GetComponent<Renderer>().material = new Material(Shader.Find("Battlehub/RTGizmos/Handles")) { color = new Color(1f, 0.1f, 0.1f, 0.3f) };
            colliderSphere.GetComponent<Renderer>().enabled = false;

            colliderRigidBody = colliderSphere.AddComponent<Rigidbody>();
            colliderRigidBody.useGravity = false;

            colliderTrigger = colliderRigidBody.gameObject.AddComponent<STriggerCollide>();
        }

        public AwasFluidCollisionTrigger(Transform parent, Vector3 positionOffset, float localScale = 0.04f) : this(localScale) {
            colliderSphere.transform.SetParent(parent);
            colliderSphere.transform.SetPositionAndRotation(parent.position + positionOffset, parent.rotation);
        }

        public void OnTriggerOnce(Atom selfAtom, Action<Collider> triggerHandler) {
            bool collided = false;
            colliderTrigger.OnTrigger += (object sender, TriggerEventArgs eventArgs) => {
                if (collided || selfAtom == eventArgs.collider.gameObject.GetComponentInParent<Atom>()) {
                    return;
                }
                if (eventArgs.evtType == "Entered") {
                    collided = true;
                    triggerHandler(eventArgs.collider);
                }
            };
        }

        public void OnTrigger(Atom selfAtom, Action<Collider> triggerHandler) {
            colliderTrigger.OnTrigger += (object sender, TriggerEventArgs eventArgs) => {
                if (selfAtom == eventArgs.collider.gameObject.GetComponentInParent<Atom>()) {
                    return;
                }
                if (eventArgs.evtType == "Entered") {
                    triggerHandler(eventArgs.collider);
                }
            };
        }

        public void SetParent(Transform newParent) {
            colliderSphere.transform.SetParent(newParent);
        }

        public Transform GetTransform() {
            return colliderSphere.transform;
        }

        public void SetTransformWithOffset(Transform source, Vector3 positionOffset) {
            colliderSphere.transform.SetPositionAndRotation(source.position + positionOffset, source.rotation);
        }

        public void SetVelocity(Vector3 velocity) {
            colliderRigidBody.velocity = velocity;
        }

        public void SetGravity(bool gravity) {
            colliderRigidBody.useGravity = gravity;
        }

        public void SetKinematic(bool kinematic) {
            colliderRigidBody.isKinematic = kinematic;
        }

        public void Stop() {
            colliderSphere.GetComponent<SphereCollider>().enabled = false;
            colliderRigidBody.velocity = Vector3.zero;
            colliderRigidBody.useGravity = false;
        }

        public void SetCollideEnabled(bool enabled) {
            colliderRigidBody.detectCollisions = enabled;
        }

        public void Remove() {
            if (colliderRigidBody != null) {
                UnityEngine.Object.Destroy(colliderRigidBody);
                colliderRigidBody = null;
            }
            if (colliderSphere != null) {
                UnityEngine.Object.Destroy(colliderSphere);
                colliderSphere = null;
            }
        }

        public void Debug() {
            colliderSphere.GetComponent<Renderer>().enabled = true;
        }
    }

    internal class AtomRigidBodyCollisionTrigger {
        private GameObject colliderSphere;
        private Atom containingAtom;
        public STriggerCollide colliderTrigger;
        private readonly Debounce triggerDebounce;

        public AtomRigidBodyCollisionTrigger(float localScale, Rigidbody colliderRigidBody, string name = "JizzCollider", float debounceDuration = 1f) {
            triggerDebounce = new Debounce(debounceDuration);

            colliderSphere = GameObject.CreatePrimitive(PrimitiveType.Sphere);
            colliderSphere.name = name;
            colliderSphere.transform.localScale = new Vector3(localScale, localScale, localScale);
            colliderSphere.GetComponent<SphereCollider>().isTrigger = true;
            colliderSphere.transform.SetParent(colliderRigidBody.gameObject.transform, false);

            colliderSphere.GetComponent<Renderer>().material = new Material(Shader.Find("Battlehub/RTGizmos/Handles")) { color = new Color(1f, 0.1f, 0.1f, 0.3f) };
            colliderSphere.GetComponent<Renderer>().enabled = false;

            containingAtom = colliderRigidBody.gameObject.GetComponentInParent<Atom>();
            colliderTrigger = colliderRigidBody.gameObject.AddComponent<STriggerCollide>();
        }

        public void OnTriggerOnce(bool allowSelfTrigger, Action<Collider> triggerHandler) {
            bool collided = false;
            colliderTrigger.OnTrigger += (object sender, TriggerEventArgs eventArgs) => {
                if (collided) return;
                if (!allowSelfTrigger && IsTriggeredBySelf(eventArgs.collider)) return;
                if (ColliderTypeInvalid(eventArgs.collider)) return;
                collided = true;
                triggerHandler(eventArgs.collider);
            };
        }

        public void OnTrigger(bool allowSelfTrigger, Action<Collider> triggerHandler) {
            colliderTrigger.OnTrigger += (object sender, TriggerEventArgs eventArgs) => {
                if (eventArgs.collider.name == "JizzBlip" || eventArgs.collider.name == "JizzCollider") {
                    triggerHandler(eventArgs.collider);
                    return;
                }
                if (!allowSelfTrigger && IsTriggeredBySelf(eventArgs.collider)) return;
                if (ColliderTypeInvalid(eventArgs.collider)) return;
                if (triggerDebounce.IsNotDebounced()) return;
                triggerHandler(eventArgs.collider);
            };
        }

        public void OnTrigger(bool allowSelfTrigger, List<string> colliderWhitelist, Action<Collider> triggerHandler) {
            colliderTrigger.OnTrigger += (object sender, TriggerEventArgs eventArgs) => {
                if (eventArgs.collider.name == "JizzBlip") {
                    triggerHandler(eventArgs.collider);
                    return;
                }
                if (!allowSelfTrigger && IsTriggeredBySelf(eventArgs.collider)) return;
                if (ColliderTypeInvalid(eventArgs.collider)) return;
                if (!ColliderTypeWhitelisted(eventArgs.collider, colliderWhitelist)) return;
                if (triggerDebounce.IsNotDebounced()) return;
                triggerHandler(eventArgs.collider);
            };
        }

        private bool IsTriggeredBySelf(Collider collider) {
            return containingAtom == collider.gameObject.GetComponentInParent<Atom>();
        }

        private bool ColliderTypeInvalid(Collider collider) {
            if (collider != null) {
                if (collider.name.Contains("PhysicsMesh")) return true;
                if (collider.name.Contains("AutoCollider")) return true;
                if (collider.name.Contains("HardCollider")) return true;
                if (collider.name.Contains("_Joint")) return true;
            }
            return false;
        }

        private bool ColliderTypeWhitelisted(Collider collider, List<string> colliderWhitelist) {
            bool whitelisted = false;
            if (collider != null) {
                for (int i = 0; i < colliderWhitelist.Count; i++) {
                    if (collider.name.Contains(colliderWhitelist[i])) {
                        whitelisted = true;
                        break;
                    }
                }
            }
            //if (!whitelisted) Logger.Debug($"{colliderSphere.name} excluding {collider.name}");
            return whitelisted;
        }

        public void SetTransformWithOffset(Transform source, Vector3 positionOffset) {
            colliderSphere.transform.SetPositionAndRotation(source.position + positionOffset, source.rotation);
        }

        public void Remove() {
            if (colliderTrigger != null) {
                UnityEngine.Object.Destroy(colliderTrigger);
                colliderTrigger.Clear();
            }
            if (colliderSphere != null) {
                UnityEngine.Object.Destroy(colliderSphere);
                colliderSphere = null;
            }
        }

        public void SetEnabled(bool enabled) {
            colliderTrigger.enabled = enabled;
        }

        public void Debug(bool enabled = true) {
            colliderSphere.GetComponent<Renderer>().enabled = enabled;
        }

        private class Debounce {

            private float timer = 0;
            private float threshold = 1f;
            public Debounce(float threshold) {
                this.threshold = threshold;
            }

            public bool IsNotDebounced() {

                bool debounced = true;
                if (Time.time - timer > threshold) {
                    debounced = false;
                    timer = Time.time;
                }
                return debounced;
            }
        }
    }

    internal class AtomRigidBodyCollider {
        private GameObject colliderSphere;

        public AtomRigidBodyCollider(float localScale, Rigidbody colliderRigidBody, string name = "JizzCollider") {

            colliderSphere = GameObject.CreatePrimitive(PrimitiveType.Sphere);
            colliderSphere.name = name;
            colliderSphere.transform.localScale = new Vector3(localScale, localScale, localScale);
            colliderSphere.GetComponent<SphereCollider>().isTrigger = true;
            colliderSphere.transform.SetParent(colliderRigidBody.gameObject.transform, false);

            colliderSphere.GetComponent<Renderer>().material = new Material(Shader.Find("Battlehub/RTGizmos/Handles")) { color = new Color(1f, 0.1f, 0.1f, 0.3f) };
            colliderSphere.GetComponent<Renderer>().enabled = false;
        }

        public void SetTransformWithOffset(Transform source, Vector3 positionOffset) {
            colliderSphere.transform.SetPositionAndRotation(source.position + positionOffset, source.rotation);
        }

        public void Remove() {
            if (colliderSphere != null) {
                UnityEngine.Object.Destroy(colliderSphere);
                colliderSphere = null;
            }
        }

        public void Debug() {
            colliderSphere.GetComponent<Renderer>().enabled = true;
        }
    }
}
