﻿using System;
using System.Linq;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using SimpleJSON;

namespace AwasAutoJuice {

    internal class AwasAutoJuiceUi {
        private MVRScript mvrScript;
        private readonly AwasActionConfigurationRepository configurationRepository;
        private string currentActionSelected;
        private bool showAdvancedSettings;
        //private JSONStorableBool autoFluidTriggersStorable;

        // "Just enough" settings
        private UIDynamicTextField settingsLabel;        
        private UIDynamicSlider maxDuration;
        private UIDynamicSlider visibleAlpha;
        private UIDynamicSlider friction;
        private UIDynamicSlider dripFriction;
        private UIDynamicSlider size;
        private UIDynamicSlider dripSize;
        private UIDynamicSlider stiffness;
        private UIDynamicSlider dripStiffness;

        private UIDynamicToggle advancedToggle;
        // Advanced settings
        private UIDynamicSlider compression;
        private UIDynamicSlider dripCompression;
        private UIDynamicSlider jointStrength;
        private UIDynamicSlider dripJointStrength;
        private UIDynamicSlider dripCollisionRadius;
        private UIDynamicSlider dripCollisionPower;
        
        // Don't mess with these
        private UIDynamicSlider skinScale;
        private UIDynamicSlider dripSkinScale;

        public AwasAutoJuiceUi(MVRScript mvrScript, AwasActionConfigurationRepository configurationRepository) {
            this.mvrScript = mvrScript;
            this.configurationRepository = configurationRepository;
            currentActionSelected = configurationRepository.actionNames[0];
            CreateActionNamesSelector(configurationRepository.actionNames);

            foreach (string actionName in configurationRepository.actionNames) {
                AwasActionConfiguration aConfig = configurationRepository.Get(actionName);
                aConfig.ui.RegisterActionConfigurationSettings(mvrScript);
            }

            AwasActionConfiguration actionConfig = configurationRepository.Get(configurationRepository.actionNames[0]);
            CreateActionConfigUi(actionConfig);
        }

        public JSONClass GetJSONForCurrentSelectedActionConfig() {
            return configurationRepository.Get(currentActionSelected).GetJSON();
        }

        public void UpdateCurrentlySelectedActionUi() {
            if (currentActionSelected != null) {
                UpdateUiForSelectedAction(configurationRepository.Get(currentActionSelected));
            }
        }

        public UIDynamicToggle CreateMainUiToggle(string name, bool initialValue, bool rightSide, Action<bool> callback) {
            // TODO this is a bit dodge. This general method shouldn't store a specific storable. Refactor.
            JSONStorableBool storable = new JSONStorableBool(name, initialValue);
            var toggle = mvrScript.CreateToggle(storable, rightSide);
            toggle.toggle.onValueChanged.AddListener(newValue =>  callback(newValue));
            return toggle;
        }

        public UIDynamicButton CreateMainUIButton(string name, bool rightStide, Action callback) {
            var button = mvrScript.CreateButton(name, rightStide);
            button.button.onClick.AddListener(() => callback());
            return button;
        }

        private void CreateActionConfigUi(AwasActionConfiguration actionConfig) {
            AwasActionConfigStorables configUi = actionConfig.ui;

            settingsLabel = CreateLabel($"{actionConfig.displayName} settings", true);
            maxDuration = CreateSlider(configUi.maxDurationStorable, true);
            visibleAlpha = CreateSlider(configUi.visibleAlphaStorable, true);
            
            friction = CreateSlider(configUi.frictionStorable, true);
            friction.valueFormat = "F3";
            
            dripFriction = CreateSlider(configUi.dripFrictionStorable, true);
            dripFriction.valueFormat = "F3";
            
            size = CreateSlider(configUi.sizeStorable, true);
            size.valueFormat = "F3";
            
            dripSize = CreateSlider(configUi.dripSizeStorable, true);
            dripSize.valueFormat = "F3";
            
            stiffness = CreateSlider(configUi.stiffnessStorable, true);
            stiffness.valueFormat = "F3";
            
            dripStiffness = CreateSlider(configUi.dripStiffnessStorable, true);
            dripStiffness.valueFormat = "F3";

            advancedToggle = CreateToggle("Show advanced settings", true);
            advancedToggle.toggle.onValueChanged.AddListener(isAdvanced => {
                showAdvancedSettings = isAdvanced;
                UpdateUiForSelectedAction(configurationRepository.Get(currentActionSelected));
            });
            
            compression = CreateSlider(configUi.compressionStorable, true);
            compression.valueFormat = "F3";
                
            dripCompression = CreateSlider(configUi.dripCompressionStorable, true);
            dripCompression.valueFormat = "F3";
                
            jointStrength = CreateSlider(configUi.jointStrengthStorable, true);
            jointStrength.valueFormat = "F3";                
                
            dripJointStrength = CreateSlider(configUi.dripJointStrengthStorable, true);
            dripJointStrength.valueFormat = "F3";                
                
            dripCollisionRadius = CreateSlider(configUi.dripCollisionRadiusStorable, true);
            dripCollisionRadius.valueFormat = "F3";                
                
            dripCollisionPower = CreateSlider(configUi.dripCollisionPowerStorable, true);
            dripCollisionPower.valueFormat = "F3";                
                
            skinScale = CreateSlider(configUi.skinScaleStorable, true);
            skinScale.valueFormat = "F3";                
                
            dripSkinScale = CreateSlider(configUi.dripSkinScaleStorable, true);
            dripSkinScale.valueFormat = "F3";            

            UpdateUiForSelectedAction(actionConfig);
        }

        private void UpdateUiForSelectedAction(AwasActionConfiguration actionConfig) {
            settingsLabel.text = $"{actionConfig.displayName} settings";

            maxDuration.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.maxDurationStorable.slider = maxDuration.slider;
            maxDuration.slider.value = actionConfig.maxDuration;
            maxDuration.slider.onValueChanged.AddListener(newValue => actionConfig.maxDuration = newValue);
            maxDuration.label = "Action duration";
            SetSliderEnabled(maxDuration, actionConfig.isDrip);

            visibleAlpha.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.visibleAlphaStorable.slider = visibleAlpha.slider;
            visibleAlpha.slider.value = actionConfig.visibleAlpha;
            visibleAlpha.slider.onValueChanged.AddListener(newValue => actionConfig.visibleAlpha = newValue);
            visibleAlpha.label = "Visible alpha";

            friction.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.frictionStorable.slider = friction.slider;
            friction.slider.value = actionConfig.friction;
            friction.slider.onValueChanged.AddListener(newValue => actionConfig.friction = newValue);
            friction.label = "Fluid friction";
            SetSliderEnabled(friction, actionConfig.isSim);

            dripFriction.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.dripFrictionStorable.slider = dripFriction.slider;
            dripFriction.slider.value = actionConfig.dripFriction;
            dripFriction.slider.onValueChanged.AddListener(newValue => actionConfig.dripFriction = newValue);
            dripFriction.label = "Drip friction";
            SetSliderEnabled(dripFriction, actionConfig.isDrip);

            size.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.sizeStorable.slider = size.slider;
            size.slider.value = actionConfig.size;
            size.slider.onValueChanged.AddListener(newValue => actionConfig.size = newValue);
            size.label = "Fluid size";
            SetSliderEnabled(size, actionConfig.isSim);

            dripSize.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.dripSizeStorable.slider = dripSize.slider;
            dripSize.slider.value = actionConfig.dripSize;
            dripSize.slider.onValueChanged.AddListener(newValue => actionConfig.dripSize = newValue);
            dripSize.label = "Drip size";
            dripSize.enabled = actionConfig.isDrip;
            SetSliderEnabled(dripSize, actionConfig.isDrip);

            stiffness.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.stiffnessStorable.slider = stiffness.slider;
            stiffness.slider.value = actionConfig.stiffness;
            stiffness.slider.onValueChanged.AddListener(newValue => actionConfig.stiffness = newValue);
            stiffness.label = "Fluid stiffness";
            SetSliderEnabled(stiffness, actionConfig.isSim);

            dripStiffness.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.dripStiffnessStorable.slider = dripStiffness.slider;
            dripStiffness.slider.value = actionConfig.dripStiffness;
            dripStiffness.slider.onValueChanged.AddListener(newValue => actionConfig.dripStiffness = newValue);
            dripStiffness.label = "Drip stiffness";
            SetSliderEnabled(dripStiffness, actionConfig.isDrip);

            compression.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.compressionStorable.slider = compression.slider;
            compression.slider.value = actionConfig.compression;
            compression.slider.onValueChanged.AddListener(newValue => actionConfig.compression  = newValue);
            compression.label = "Compression";
            SetSliderEnabled(compression, actionConfig.isSim);

            dripCompression.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.dripCompressionStorable.slider = dripCompression.slider;
            dripCompression.slider.value = actionConfig.dripCompression;
            dripCompression.slider.onValueChanged.AddListener(newValue => actionConfig.dripCompression = newValue);
            dripCompression.label = "Drip compression";
            SetSliderEnabled(dripCompression, actionConfig.isDrip);

            jointStrength.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.jointStrengthStorable.slider = jointStrength.slider;
            jointStrength.slider.value = actionConfig.jointStrength;
            jointStrength.slider.onValueChanged.AddListener(newValue => actionConfig.jointStrength = newValue);
            jointStrength.label = "Joint strength";
            SetSliderEnabled(jointStrength, actionConfig.isSim);

            dripJointStrength.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.dripJointStrengthStorable.slider = dripJointStrength.slider;
            dripJointStrength.slider.value = actionConfig.dripJointStrength;
            dripJointStrength.slider.onValueChanged.AddListener(newValue => actionConfig.dripJointStrength = newValue);
            dripJointStrength.label = "Drip joint strength";
            SetSliderEnabled(dripJointStrength, actionConfig.isDrip);

            dripCollisionRadius.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.dripCollisionRadiusStorable.slider = dripCollisionRadius.slider;
            dripCollisionRadius.slider.value = actionConfig.dripCollisionRadius;
            dripCollisionRadius.slider.onValueChanged.AddListener(newValue => actionConfig.dripCollisionRadius = newValue);
            dripCollisionRadius.label = "Drip collision radius";
            SetSliderEnabled(dripCollisionRadius, actionConfig.isSim);

            dripCollisionPower.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.dripCollisionPowerStorable.slider = dripCollisionPower.slider;
            dripCollisionPower.slider.value = actionConfig.dripCollisionPower;
            dripCollisionPower.slider.onValueChanged.AddListener(newValue => actionConfig.dripCollisionPower = newValue);
            dripCollisionPower.label = "Drip collision power";
            SetSliderEnabled(dripCollisionPower, actionConfig.isSim);

            skinScale.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.skinScaleStorable.slider = skinScale.slider;
            skinScale.slider.value = actionConfig.skinScale;
            skinScale.slider.onValueChanged.AddListener(newValue => actionConfig.skinScale = newValue);
            skinScale.label = "Skin scale";
            skinScale.gameObject.SetActive(false);

            dripSkinScale.slider.onValueChanged.RemoveAllListeners();
            actionConfig.ui.dripSkinScaleStorable.slider = dripSkinScale.slider;
            dripSkinScale.slider.value = actionConfig.dripSkinScale;
            dripSkinScale.slider.onValueChanged.AddListener(newValue => actionConfig.dripSkinScale = newValue);
            dripSkinScale.label = "Drip skin scale";
            dripSkinScale.gameObject.SetActive(false);

            var advancedSliders = new List<UIDynamicSlider> { compression, dripCompression, jointStrength, dripJointStrength, dripCollisionRadius, dripCollisionPower };
            if (RunMode.isDev) {
                advancedSliders.Add(skinScale);
                advancedSliders.Add(dripSkinScale);
            }
            SetAdvancedSlidersColor(advancedSliders);
            SetAdvancedSlidersVisible(showAdvancedSettings, advancedSliders);
        }

        private void CreateActionNamesSelector(List<string> actionNames) {
            var displayActionNames = new List<string>();
            actionNames.ForEach(actionName => displayActionNames.Add(configurationRepository.Get(actionName).displayName));

            JSONStorableStringChooser storableStringChooser = new JSONStorableStringChooser("Action", actionNames, displayActionNames, actionNames[0], "Action") {
                isRestorable = false, isStorable = false
            };
            storableStringChooser.setCallbackFunction += (string actionChoiceName) => {
                UpdateUiForSelectedAction(configurationRepository.Get(actionChoiceName));
                currentActionSelected = actionChoiceName;
            };
            UIDynamicPopup scrollablePopup = mvrScript.CreateScrollablePopup(storableStringChooser, false);
            scrollablePopup.labelTextColor = Color.black;
        }

        private void SetSliderEnabled(UIDynamicSlider slider, bool enabled) {
            slider.slider.interactable = slider.quickButtonsEnabled = slider.defaultButtonEnabled = enabled;
            slider.labelText.color = enabled ? Color.black : Color.grey;
            slider.sliderValueTextFromFloat.UIInputField.interactable = enabled;
            slider.sliderValueTextFromFloat.UIInputField.textComponent.color = enabled ? Color.black : Color.grey;
        }

        private void SetAdvancedSlidersVisible(bool enabled, List<UIDynamicSlider> sliders) {
            foreach (UIDynamicSlider slider in sliders) {
                slider.gameObject.SetActive(enabled);
            }
        }

        private void SetAdvancedSlidersColor(List<UIDynamicSlider> sliders) {
            foreach (UIDynamicSlider slider in sliders) {
                var images = slider.gameObject.transform.GetComponentsInChildren<Image>();
                var fillImage = images.First(image => image.name == "Panel");
                fillImage.color = new Color(1f, 0.95f, 0.95f);
            }
        }

        private UIDynamicSlider CreateSlider(JSONStorableFloat jSONStorableFloat, bool rightSide = false) {
            UIDynamicSlider slider = mvrScript.CreateSlider(jSONStorableFloat, rightSide);
            slider.sliderValueTextFromFloat.UIInputField.textComponent.fontSize = slider.labelText.fontSize = 20;
            return slider;
        }

        private UIDynamicTextField CreateLabel(string uiLabel, bool rightSide = false) {
            JSONStorableString tempStorable = new JSONStorableString("", uiLabel) { isRestorable = false, isStorable = false };
            UIDynamicTextField labelTextField = mvrScript.CreateTextField(tempStorable, rightSide);
            labelTextField.backgroundColor = new Color(1, 1, 1, 0.25f);
            labelTextField.UItext.GetComponent<RectTransform>().anchoredPosition = new Vector2(10f, -10f);
            ScrollRect[] scrollRects = labelTextField.UItext.transform.GetComponentsInParent<ScrollRect>(true);
            for (int i = 0; i < scrollRects.Length; i++) {
                scrollRects[i].horizontal = scrollRects[i].vertical = false;
            }
            LayoutElement labelLayout = labelTextField.GetComponent<LayoutElement>();
            labelLayout.minHeight = labelLayout.preferredHeight = labelTextField.height = 50f;
            return labelTextField;
        }

        private UIDynamicToggle CreateToggle(string toggleLabel, bool rightSide = false) {
            return mvrScript.CreateToggle(new JSONStorableBool(toggleLabel, false) { isRestorable = false, isStorable = false }, rightSide);
        }
    }

    internal class AwasActionConfigStorables {

        private readonly AwasActionConfiguration actionConfig;
        public JSONStorableFloat maxDurationStorable;
        public JSONStorableFloat visibleAlphaStorable;
        public JSONStorableFloat frictionStorable;
        public JSONStorableFloat dripFrictionStorable;
        public JSONStorableFloat dripCompressionStorable;
        public JSONStorableFloat sizeStorable;
        public JSONStorableFloat dripSizeStorable;
        public JSONStorableFloat stiffnessStorable;
        public JSONStorableFloat dripStiffnessStorable;

        public JSONStorableFloat compressionStorable;
        public JSONStorableFloat jointStrengthStorable;
        public JSONStorableFloat dripJointStrengthStorable;
        public JSONStorableFloat dripCollisionRadiusStorable;
        public JSONStorableFloat dripCollisionPowerStorable;
        public JSONStorableFloat skinScaleStorable;
        public JSONStorableFloat dripSkinScaleStorable;

        public AwasActionConfigStorables(AwasActionConfiguration actionConfig) {
            this.actionConfig = actionConfig;
        }

        public void RegisterActionConfigurationSettings(MVRScript mvrScript) {
            mvrScript.RegisterString(new JSONStorableString($"--- {actionConfig.name} ---", "") { isRestorable = false, isStorable = false });
            maxDurationStorable = CreateFloatStorable($"{actionConfig.name} Max Duration", actionConfig.maxDuration, ActionSettingsRange.duration, newValue => actionConfig.maxDuration = newValue);
            visibleAlphaStorable = CreateFloatStorable($"{actionConfig.name} Visible Alpha", actionConfig.visibleAlpha, ActionSettingsRange.alpha, newValue => actionConfig.visibleAlpha = newValue);
            frictionStorable = CreateFloatStorable($"{actionConfig.name} Friction", actionConfig.friction, ActionSettingsRange.zeroToOne, newValue => actionConfig.friction = newValue);
            dripFrictionStorable = CreateFloatStorable($"{actionConfig.name} Drip Friction", actionConfig.dripFriction, ActionSettingsRange.zeroToOne, newValue => actionConfig.dripFriction = newValue);
            sizeStorable = CreateFloatStorable($"{actionConfig.name} Fluid Size", actionConfig.size, ActionSettingsRange.size, newValue => actionConfig.size = newValue);
            dripSizeStorable = CreateFloatStorable($"{actionConfig.name} Drip Size", actionConfig.dripSize, ActionSettingsRange.size, newValue => actionConfig.dripSize = newValue);
            stiffnessStorable = CreateFloatStorable($"{actionConfig.name} Fluid Stiffness", actionConfig.stiffness, ActionSettingsRange.zeroToOne, newValue => actionConfig.stiffness = newValue);
            dripStiffnessStorable = CreateFloatStorable($"{actionConfig.name} Drip Stiffness", actionConfig.dripStiffness, ActionSettingsRange.zeroToOne, newValue => actionConfig.dripStiffness = newValue);

            compressionStorable = CreateFloatStorable($"{actionConfig.name} Compression", actionConfig.compression, ActionSettingsRange.nonZero, newValue => actionConfig.compression = newValue);
            dripCompressionStorable = CreateFloatStorable($"{actionConfig.name} Drip Compression", actionConfig.dripCompression, ActionSettingsRange.nonZero, newValue => actionConfig.dripCompression = newValue);
            jointStrengthStorable = CreateFloatStorable($"{actionConfig.name} Joint Strength", actionConfig.jointStrength, ActionSettingsRange.zeroToOne, newValue => actionConfig.jointStrength = newValue);
            dripJointStrengthStorable = CreateFloatStorable($"{actionConfig.name} Drip Joint Strength", actionConfig.dripJointStrength, ActionSettingsRange.zeroToOne, newValue => actionConfig.dripJointStrength = newValue);
            dripCollisionRadiusStorable = CreateFloatStorable($"{actionConfig.name} Drip Collision Radius", actionConfig.dripCollisionRadius, ActionSettingsRange.nonZero, newValue => actionConfig.dripCollisionRadius = newValue);
            dripCollisionPowerStorable = CreateFloatStorable($"{actionConfig.name} Drip Collision Power", actionConfig.dripCollisionPower, ActionSettingsRange.zeroToOne, newValue => actionConfig.dripCollisionPower = newValue);
            skinScaleStorable = CreateFloatStorable($"{actionConfig.name} Skin Scale", actionConfig.skinScale, ActionSettingsRange.size, newValue => actionConfig.skinScale = newValue);
            dripSkinScaleStorable = CreateFloatStorable($"{actionConfig.name} Drip Skin Scale", actionConfig.dripSkinScale, ActionSettingsRange.size, newValue => actionConfig.dripSkinScale = newValue);

            mvrScript.RegisterFloat(visibleAlphaStorable);
            if (actionConfig.isSim) {
                mvrScript.RegisterFloat(frictionStorable);
                mvrScript.RegisterFloat(sizeStorable);
                mvrScript.RegisterFloat(stiffnessStorable);
            }
            if (actionConfig.isDrip) {
                mvrScript.RegisterFloat(maxDurationStorable);
                mvrScript.RegisterFloat(dripFrictionStorable);
                mvrScript.RegisterFloat(dripSizeStorable);
                mvrScript.RegisterFloat(dripStiffnessStorable);
            }

            // TODO what about registering advanced storables?
        }

        public JSONStorableFloat CreateFloatStorable(string name, float start, ActionSettingsRange.Range range, Action<float> onChange) {
            return new JSONStorableFloat(name, start, (newValue) => {
                onChange(newValue);
            }, range.min, range.max) { isStorable = false, isRestorable = false };
        }
    }

}
